#ifndef lint
static char rcsid[] =
    "@(#) $Header: dbridge.c,v 1.14 87/05/10 19:45:52 leres Exp $ (LBL)";
#endif
/*
 * TCP/IP bridging using DECNET
 *
 * Copyright 1984 by Van Jacobson and Craig Leres
 *		     Real Time Systems Group
 *		     Lawrence Berkeley Laboratory
 *		     Berkeley, CA 94720
 *
 * $Log:	dbridge.c,v $
 * Revision 1.14  87/05/10  19:45:52  leres
 * Need sys/param.h for netinet/ip.h.
 * 
 * Revision 1.13  87/05/10  19:32:44  leres
 * Modify to use vmsperror().
 *
 * Revision 1.12  87/01/24  20:39:21  leres
 * Fall back on SYS$NODE if sys$getsyiw() fails us.
 *
 * Revision 1.11  86/12/16  15:39:12  leres
 * Use setsockopt() to increase socket rcvbuf size.
 *
 * Revision 1.10  86/08/31  17:01:40  leres
 * Use sys$getsyiw() to get local decnet hostname. Translating
 * SYS$NODE is too unreliable.
 */
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <eunice/eunice.h>
#include <vms/iodef.h>
#include <vms/msgdef.h>
#ifdef notdef
#include <vms/ssdef.h>
#else
#define	SS$_LINKABORT	8420		/* only define used from ssdef.h */
#endif
#include <vms/dibdef.h>
#include <vms/nfbdef.h>
#include <vms/syidef.h>
#include <netdb.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
char *getenv();

#define MAXPACKET 2048		/* max. packet size */
#define MAX_HOST_NUMBER 256	/* largest decnet/inet host number we accept */
				/* should be a power of 2 */
#define	HOST_NUMBER_MASK (MAX_HOST_NUMBER-1)
#define MAX_CONNECTIONS 16	/* max active decnet connections */
#define NULL 0
static int net_mb_maxmsg = 256;	/* max decnet mailbox msg size */
static int net_mb_bufquo = 512; /* decnet mailbox buffer quota */

/* this is the event flag used for socket reads & decnet writes */
/* It really should be allocated using lib$getefn but I don't know */
/* if Eunice allocates flags using lib$getefn and, if it doesn't, */
/* we'll end up with conflicting efn assignments. */
#define background_efn	11	/* efn for task level read/writes */
/* this event flag is used for ast level qiow's.  Should also be */
/* allocated with lib$getefn. */
#define ast_efn		12

/* the following is a 1 min. timeout for decnet writes. */
static int write_timeout[] = { -60*10*1000*1000, -1 };

struct ip_packet {
	struct ip ip_hdr;
	char	ip_dat[MAXPACKET];
};
struct net_con {
	struct	net_con *nc_next;	/* next active connection */
	long	nc_unit;		/* decnet "net:" device unit # */
	u_short	nc_channel;		/* vms i/o channel for connection */
	short	nc_host;		/* host number for connection */
	u_short	nc_iosb[4];		/* i/o status for decnet i/o */
	struct	ip_packet nc_packet;	/* buffer for i/o on connection */
};

static struct net_con *net_con_list = (struct net_con *)NULL;
static int active_con = 0;

static struct net_con * inet_host[MAX_HOST_NUMBER+1];
static struct in_addr local_host;

static struct sockaddr sock_addr = {PF_RAWPACKET};
static struct sockaddr_in *from_addr;
static struct {
	unsigned short from_len;
	char	from_dat[128];
} sock_from;


struct dnmb_msg {			/* decnet mailbox msg structure */
	unsigned short nmb_type;
	unsigned short nmb_unit;
	unsigned char nmb_text[128];
};

static int User_Space[2] = {0x0, 0x7fffffff};

static u_short sock_iosb[4];
static u_short sock_chan;

struct ip_packet sock_packet;

extern int errno;
extern int vmserrno;

int sock_ast();
char *paddr();
struct net_con *find_net_con();
struct net_con *new_decnet_connection();
struct net_con *connect_to_host();
struct net_con *connect_to_host_by_name();
struct in_addr *get_host_addr();
int net_mb_ast();

struct Descriptor {
	long	d_length;
	char	*d_address;
};

static struct dnmb_msg net_mb_msg;	/* buffer for decnet msgs */
static u_short net_mb_chan;		/* mailbox channel from _NET: */
static u_short net_mb_iosb[4];
static u_short net_ctrl_chan;		/* control channel to _NET: */
static u_short net_ctrl_iosb[4];
static struct Descriptor netdev_desc;	/* descriptor for "_NET:" */
static struct Descriptor net_mbx_desc;	/* descriptor for net mailbox */

static int debug = 0;

/* convert an internet host number into an index into the network
 * connection table.  The result has to be less than MAX_HOST_NUMBER.
 * /etc/hosts should be set up so the largest network number used
 * (excluding the subnet number) is < MAX_HOST_NUMBER).
 */
#define	NC_INDEX(host)	(ntohl(host.s_addr) & HOST_NUMBER_MASK)

/* macros to make the vms qio's look cleaner. */

#define SOCK_READ \
	sys$qiow(background_efn,sock_chan,IO$_READVBLK,sock_iosb, \
		0, 0, &sock_packet, sizeof sock_packet, 0, \
		&sock_from, sizeof sock_from, 0 )

#define SOCK_WRITE(buf,bufsize,iosb) \
	sys$qiow(ast_efn,sock_chan,IO$_WRITEVBLK,iosb,0,0,\
		buf,bufsize,0,&sock_addr,sizeof sock_addr,0)

#define DECNET_READ(nc) \
	sys$qio(_$ASYNCHRONOUS_EFN,nc->nc_channel,IO$_READVBLK,\
		nc->nc_iosb,decnet_ast,nc,\
		&nc->nc_packet,sizeof nc->nc_packet,0,0,0,0)

#define DECNET_WRITE(nc,buf,bufsize,iosb) \
	sys$qiow(background_efn,nc->nc_channel,IO$_WRITEVBLK,iosb,0,0,\
		buf,bufsize,0,0,0,0)

#define DNMB_READ \
	sys$qio(_$ASYNCHRONOUS_EFN,net_mb_chan,IO$_READVBLK,net_mb_iosb, \
		net_mb_ast, 0, &net_mb_msg, sizeof net_mb_msg, 0, 0, 0, 0 )

#define QIO(chan,code,iosb,p1,p2,p3,p4) \
	sys$qiow(ast_efn,chan,code,iosb,0,0,p1,p2,p3,p4,0,0)


/* definitions to make lint shut up about the variable arguments to sys$qio */

#ifdef lint
/*VARARGS4,ARGSUSED*/
int sys$qio(efn,chan,code,iosb)
int efn, code;
u_short chan,iosb[];
{ return(0); }
/*VARARGS4,ARGSUSED*/
int sys$qiow(efn,chan,code,iosb)
int efn, code;
u_short chan,iosb[];
{ return(0); }
/*VARARGS2,ARGSUSED*/
sprintf(dst,fmt)
char *dst;
char *fmt;
{ }
#endif


main(argc,argv)
char *argv[];
{
	register int i, host, len;
	register struct net_con *nc;
	register struct FABRAB *FabRab;
	char *usage = "usage: dbridge [-d] [-h<host>...]\n";
	int s;
	int sobufsize;

	/* Setup facility name for this program (used for opcom messages) */
	opcominit("dnet<->inet");

	from_addr = (struct sockaddr_in *) sock_from.from_dat;
	setup_decnet();

	for( i=1; i<argc; i++ ) {
		if( argv[i][0] != '-' ) {
			printf(usage);
			exit(1);
		}
		if( argv[i][1] == 'd' )
			debug++;

		else if( argv[i][1] == 'h' ) {
			if( ! connect_to_host_by_name( &(argv[i][2]) ) )
				printf("couldn't connect to host \"%s\"\n",
					&(argv[i][2]) );
		}
		else {
			printf(usage);
			exit(1);
		}
	}

	connect_to_local_host();

	/* get a raw socket handle */
	s = socket(AF_RAWPACKET, SOCK_RAW, 0);
	if (s < 0) {
		perror("socket");
		exit(1);
	}

	/* make sure we have lots of buffering on the raw socket */
	sobufsize = 31744; /* (32K - 1K to avoid sign bit problems) */
	if (setsockopt(s, SOL_SOCKET, SO_RCVBUF,
	      &sobufsize, sizeof(sobufsize)) < 0) {
		perror("rcvbufsize");
		exit(1);
	}

	FabRab = FD_FAB_Pointer[s];
	sock_chan = FabRab->fab.fab$l_stv;

	i = sys$purgws(User_Space);

	/*
	 * loop forever reading packets from the socket & forwarding them
	 * to decnet.  We do this at task level so we can hang a timer
	 * on the decnet write and still be getting packets from decnet
	 * at ast level.
	 */
	while( 1 ) {
		i = SOCK_READ;
		if( ! (i & 1) )
			io_error( "socket read", (u_short) i );
		if( ! (sock_iosb[0] & 1) ) {
			io_error( "socket read", sock_iosb[0] );
		}
		len = sock_iosb[1];
#ifdef	DEBUG
		if( debug ) {
			printf("send %d bytes, dst %s, src %s", len,
			    inet_ntoa(sock_packet.ip_hdr.ip_dst),
			    inet_ntoa(sock_packet.ip_hdr.ip_src));
			if( sock_from.from_len )
				printf(", from %s",
				    inet_ntoa(from_addr->sin_addr));
			printf("\n");
		}
#endif
		/*
		 * if there's no connection open to the decnet host, try
		 * to open one.  once there's a connection, forward the
		 * packet.  If we get any errors forwarding the packet,
		 * disconnect the circuit (the errors are probably due to
		 * the other end disconnecting but we need to clean up
		 * our internal connection tables).
		 */
		host = NC_INDEX(from_addr->sin_addr);
		if( len >= sizeof(struct ip) ) {
			if( inet_host[host] ||
			    connect_to_host( from_addr->sin_addr ) ) {
				nc = inet_host[host];
				if( decnet_write( nc, &sock_packet, len ) )
					dn_disconnect( nc );
			}
		}
	}
}


/*
 * routine to do a timed write to a decnet channel.  we return
 * 0 if no errors, -1 if timeout or any error.
 */
decnet_write( nc, packet, len )
struct net_con *nc;
struct ip_packet *packet;
int len;
{
	register int i;
	u_short iosb[4];

	/*
	 * start a timer going using the same efn that we use for
	 * the write.  Either the write completion or the timer
	 * expiration should restart us.  Timer expiration will leave
	 * us with an even number (0) in the iosb, as will any i/o
	 * error.  Since we disconnect the channel on any error, we
	 * don't bother to cancel the write on a timeout.
	 */
	sys$setimr( background_efn, write_timeout, 0, 1234 );
	i = DECNET_WRITE( nc, packet, len, iosb );
	sys$cantim( 1234, 0 );
	if( (i & 1) && (iosb[0] & 1) )
		return( 0 );

	return(-1);
}



/* decnet read completion ast - feed the data to ip_input via socket */

decnet_ast( nc )
struct net_con *nc; {
	register int i, len;
	u_short iosb[4];
	int dont_block = 1;
	int block = 0;

	if( ! (nc->nc_iosb[0] & 1) ) {
		io_error( "decnet read", nc->nc_iosb[0] );
		dn_disconnect( nc );
		return;
	}
	len = nc->nc_iosb[1];
#ifdef	DEBUG
	if( debug )
		printf("rcvd %d bytes, dst %s, src %s\n", len,
		    inet_ntoa(nc->nc_packet.ip_hdr.ip_dst),
		    inet_ntoa(nc->nc_packet.ip_hdr.ip_src));
#endif
	/*
	 * send the data back through ip.  do the write in non-blocking
	 * mode so we'll just drop the packet if the buffer is clogged.
	 */
	sys$qiow(ast_efn,sock_chan,IO$_IOCTL,iosb,0,0,
		 FIONBIO,&dont_block,0,0,0,0);
	SOCK_WRITE( &nc->nc_packet, len, iosb );
	sys$qiow(ast_efn,sock_chan,IO$_IOCTL,iosb,0,0,
		 FIONBIO,&block,0,0,0,0);

	i = DECNET_READ(nc);
	if( ! (i & 1) )
		dn_disconnect( nc );
}


/* DECNET Network Management Routines */

/*
 * Set up our Decnet connection & make ourself the server for msgs
 * to DN2IN.
 */
setup_decnet()
{
	register int i;
	struct Descriptor nwnamedesc, nfbdesc;
	char nfb[5];

	nfb[0] = NFB$C_DECLNAME;
	nfb[1] = nfb[2] = nfb[3] = nfb[4] = 0;

	/* set up all the "descriptors" we'll need to talk to decnet */
	netdev_desc.d_address = "_NET:";
	netdev_desc.d_length = strlen( netdev_desc.d_address );

	net_mbx_desc.d_address = "DN2IN_MBX";
	net_mbx_desc.d_length = strlen( net_mbx_desc.d_address );

	nwnamedesc.d_address = "DN2IN";
	nwnamedesc.d_length = strlen( nwnamedesc.d_address );

	nfbdesc.d_address = nfb;
	nfbdesc.d_length = sizeof nfb;

	/* Set up a control msg connection to Decnet */
	i = sys$crembx( 0, &net_mb_chan, net_mb_maxmsg, net_mb_bufquo,
			0, 3, &net_mbx_desc );
	if( ! (i & 1) )
		qio_error( "couldnt create decnet mailbox.", i );
	i = sys$assign( &netdev_desc, &net_ctrl_chan, 3, &net_mbx_desc );
	if( ! (i & 1) )
		qio_error( "_NET: sys$assign error", i );

	/* declare ourselves as object DN2IN */
	i = QIO( net_ctrl_chan, IO$_ACPCONTROL, net_ctrl_iosb, &nfbdesc,
		 &nwnamedesc, 0, 0 );
	if( ! (i & 1) )
		qio_error( "decnet acpcontrol", i );
	if( ! (net_ctrl_iosb[0] & 1) ) {
		io_error( "decnet acpcontrol", net_ctrl_iosb[0] );
		exit(1);
	}

	/* start up the intial read on the Decnet mailbox */
	i = DNMB_READ;
	if( ! (i & 1) )
		qio_error( "decnet mailbox read", i );
}



/*
 * Process a message from the Decnet Mailbox (some sort of network
 * control operation like a connect or disconnect req. from our peer
 * on some other system).
 */
net_mb_ast() {
	register int i;
	register struct net_con *nc;

	if ( ! (net_mb_iosb[0] & 1) ) {
		io_error( "decnet mailbox read", net_mb_iosb[0] );
		exit(1);
	}

	switch( net_mb_msg.nmb_type ) {

		case MSG$_ABORT:
		case MSG$_DISCON:
		case MSG$_EXIT:
		case MSG$_PATHLOST:
		case MSG$_PROTOCOL:
		case MSG$_THIRDPARTY:
		case MSG$_TIMEOUT:
			if( nc = find_net_con( net_mb_msg.nmb_unit ) )
				dn_disconnect( nc );
			break;

		case MSG$_NETSHUT:
			/* exit */
			exit(0);

		case MSG$_INTMSG:
		case MSG$_REJECT:
		case MSG$_CONFIRM:
			/* ignore */
			break;

		case MSG$_CONNECT:
			dn_connect( net_mb_msg );
			break;

		default:
			printf( "unrecognized decnet msg type %d\n",
			    net_mb_msg.nmb_type );
			break;
	}
	/* requeue the read on the decnet mailbox. */

	i = DNMB_READ;
	if( ! (i & 1) )
		qio_error( "decnet mailbox read", i );
}


/* find a decnet connection given its unit number */

struct net_con *
find_net_con( unit )
u_short unit;
{
	register struct net_con *nc = net_con_list;

	/* find the connection associated with this unit */

	while( nc && nc->nc_unit != unit )
		nc = nc->nc_next;

	return( nc );
}



/* disconnect a decnet connection */

dn_disconnect( nc )
struct net_con *nc; {
	struct net_con *prev;

	if ( nc == 0 || inet_host[nc->nc_host] == 0 )
		return;

	/* deassign the channel to close the connection, then free the */
	/* space associated with the connection */

	sys$dassgn( nc->nc_channel );
	inet_host[nc->nc_host] = 0;
	active_con--;

	/* delete this entry from the active connections list */
	if( net_con_list == nc )
		net_con_list = nc->nc_next;
	else {
		prev = net_con_list;
		while( prev->nc_next && prev->nc_next != nc )
			prev = prev->nc_next;
		if( prev->nc_next )
			prev->nc_next = nc->nc_next;
		else {
			printf( "dn_disconnect(): couldn't find connection\n" );
			exit(1);
		}
	}
	free( nc );
}



/* accept or reject a request for a decnet connection */

dn_connect( nmsg )
struct dnmb_msg nmsg; {
	register int i;
	struct Descriptor ncb;
	struct net_con *nc;
	struct in_addr host;
	struct in_addr *hpntr;
	char dnhost[128];
	char *ch, *ch2;
	u_short iosb[4];

	/* extract the ncb from the decnet msg */

	i = nmsg.nmb_text[0] + 1;
	ncb.d_length = nmsg.nmb_text[i];
	ncb.d_address = (char *) & nmsg.nmb_text[i+1];

	/* This is a horrible mess -- what we really want to do is */
	/* to make sure we have this host in /etc/hosts before we  */
	/* accept the connection.  Unfortunately, it takes so long */
	/* to look up something in /etc/hosts that netacp would    */
	/* time out this connect req. if we stopped to look up the */
	/* host number.  So we charge ahead & make the connection  */
	/* then immediately disconnect if we don't know the host.  */

	host.s_addr = 0;
	if( !(nc = new_decnet_connection( &ncb, host ) ) ) {
		/* refuse connection */
		QIO( net_ctrl_chan, IO$_ACCESS | IO$M_ABORT,
		     iosb, 0, &ncb, 0, 0 );
		return;
	}
	/* get the decnet hostname from the ncb then try to get its */
	/* number from /etc/hosts.  if this host isn't known to us  */
	/* disconnect the connection. */

	ch = dnhost;
	ch2 = ncb.d_address;
	for( i=0; ch2[i] && ch2[i] == '_'; i++ )
		;
	for( ; ch2[i] && ch2[i] != ':'; i++ )
		*ch++ = (isupper(ch2[i])? tolower(ch2[i]) : ch2[i]);

	/* try to find <decnet-host-name>-dbridge in /etc/hosts.  if that
	 * fails, try plain <decnet-host-name>.
	 */
	strcpy(ch, "-dbridge");
	if( !(hpntr = get_host_addr(dnhost)) )
		{
		ch = rindex(dnhost, '-');
		*ch = '\0';
		hpntr = get_host_addr(dnhost);
		}
	if ( !hpntr )
		{
		printf( "connect request from unknown host \"%s\"\n", dnhost );
		dn_disconnect( nc );
		return;
		}
	/* since we now know the host number, straighten out the */
	/* host pointers in inet_host & the nc (we cant do this  */
	/* if what we have is the local_host because we need an  */
	/* outbound & an inbound connection) */
	host = *hpntr;
	nc->nc_host = (host.s_addr != local_host.s_addr)?
			NC_INDEX(host) :
			MAX_HOST_NUMBER;
	inet_host[nc->nc_host] = nc;
	inet_host[0] = (struct net_con *) 0;
}



/*
 * attempt to setup a new decnet connection (used both for both outbound
 * and inbound connection requests).
 */
struct net_con *
new_decnet_connection( ncb, host )
struct Descriptor *ncb;
struct in_addr host;
{
	struct net_con *nc;
	int i, ch_info_len;
	u_short net_chan;
	struct DIB ch_info;
	struct Descriptor ch_info_desc;

	if( active_con >= MAX_CONNECTIONS )
		return( NULL );

	/* assign a new channel to _NET: and get its unit number */

	i = sys$assign( &netdev_desc, &net_chan, 3, &net_mbx_desc );
	if( ! (i & 1) )
		qio_error( "sys$assign error %d", i );

	ch_info_desc.d_length = sizeof (struct DIB);
	ch_info_desc.d_address = (char *) &ch_info;

	i = sys$getchn( net_chan, &ch_info_len, &ch_info_desc,
			&ch_info_len, 0 );
	if( ! (i & 1) )
		qio_error( "sys$getchn error %d", i );

	/* setup a new connection descriptor */

	nc = (struct net_con *) malloc( sizeof (struct net_con) );
	if( ! nc ) {
		printf( "new_decnet_connection(): malloc failure\n" );
		exit(1);
	}
	nc->nc_channel = net_chan;
	nc->nc_unit = ch_info.dib$w_unit;
	nc->nc_host = NC_INDEX(host);
	nc->nc_next = net_con_list;
	net_con_list = nc;
	inet_host[nc->nc_host] = nc;
	active_con++;

	/* try to do the decnet connect */

	i = QIO( net_chan, IO$_ACCESS, nc->nc_iosb, 0, ncb, 0, 0 );
	if( !(i & 1) || !(nc->nc_iosb[0] & 1) ) {
		/* connect failure - clean up this side */
		dn_disconnect( nc );
		nc = (struct net_con *) NULL;
	}

	/* queue up the initial read on the new connection */

	if( nc ) {
		i = DECNET_READ(nc);
		if( ! (i & 1) )
			io_error( "new connection rd error", (u_short) i );
	}
	return(nc);
}


/*
 * Attempt to connect to Decnet host associated with inet host
 * number "host".  (called whenever there's an ip packet destined
 * for "host" but no open decnet connection).  Convert the host
 * number to a name then open a connection to our peer on that host.
 */
struct net_con *
connect_to_host( host )
struct in_addr host;
{
	struct Descriptor ncb;
	char ncb_dat[128];
	struct hostent *dest;
	register char *hp, *p;
	register int i;

	/* get the host name associated with this number */
	dest = gethostbyaddr( &host, sizeof host, AF_INET );
	if( dest == 0 )
		return( 0 );

	/* try to find a name of the form xxx-dbridge in the host name
	 * or alias name fields.  If found, use the xxx as the decnet host
	 * name.  If not found, use the host name as the decnet host name.
	 */
	hp = dest->h_name;
	if (!(p = rindex(hp, '-')) || strcmp(p, "-dbridge") )
		{
		while( hp = *(dest->h_aliases++) )
			if ((p = rindex(hp, '-')) && strcmp(p, "-dbridge") == 0)
				break;
		}
	if ( hp )
		*p = '\0';
	else
		hp = dest->h_name;

	sprintf( ncb_dat, "%s::\"TASK=DN2IN/xx\"", hp );
	i = strlen( ncb_dat );
	ncb_dat[i-2] = ncb_dat[i-3] = 0;
	ncb.d_length = i;
	ncb.d_address = ncb_dat;

	return( new_decnet_connection( &ncb, host ) );
}

struct net_con *
connect_to_host_by_name( hname )
char *hname;
{
	struct in_addr *d;
	struct in_addr destaddr;

	if( d = get_host_addr( hname ) ) {
		destaddr = *d;
		return( connect_to_host(destaddr) );
	}
	return( 0 );
}

struct item_list_3 {
	short i_length;				/* length of user buffer */
	short i_code;				/* request code number */
	caddr_t i_buffer;			/* buffer address */
	short *i_rslength;			/* result length address */
};

static short ret;

static struct item_list_3 itmlst[2] = {
	{ 15, SYI$_NODENAME, (caddr_t)0, &ret },
	{ 0, 0, (char *)0, (short *)0 }
};

connect_to_local_host()
{
	register char *cp;
	int status, iosb[2];
	struct in_addr *h;
	char thostb[128];

	/* Get the local decnet hostname */
	ret = 0;
	itmlst[0].i_buffer = thostb;
	status = sys$getsyiw(0, 0, 0, itmlst, iosb, 0, 0);
	if ((status & 1) != 0)
		status = iosb[0];
	if ((status & 1) == 0)
		io_error("connect_to_local_host(): getsyiw()", status);
	thostb[ret] = '\0';
	if (thostb[0] == '\0') {
		/* Fall back on SYS$NODE */
		if ((cp = getenv("SYS$NODE")) != NULL) {
			while (*cp == '_')
				cp++;
			strcpy(thostb, cp);
			if ((cp = index(thostb, ':')) != NULL)
				*cp = '\0';
		}
	}
	if (thostb[0] == '\0') {
		printf("Couldn't get local decnet hostname\n");
		exit(2);
	}
	for (cp = thostb; *cp != '\0'; cp++)
		if (isupper(*cp))
			*cp = tolower(*cp);

	/* Try to find the name with & without a trailing "-dbridge" */
	(void) strcat(thostb, "-dbridge");
	if ( !(h = get_host_addr(thostb)) ) {
		*rindex(thostb, '-') = '\0';
		h = get_host_addr(thostb);
	}
	if ( ! h ) {
		printf("Couldn't get inet host number for \"%s\"\n", thostb);
		exit(2);
	}
	local_host = *h;
	if ( ! connect_to_host(local_host) ) {
		printf("Couldn't connect to local host \"%s\"\n", thostb);
		exit(2);
	}
}

/* get the inet host number given a hostname */

struct in_addr *
get_host_addr( hname )
char *hname;
{
	struct hostent *dest;

	dest = gethostbyname( hname );
	if( dest )
		return( (struct in_addr *) dest->h_addr );

	return( 0 );
}


/* Error handling and Reporting Routines */

/*
 * qio_error - this routine is called if we get an error issuing a qio.
 * we write out the qio status then exit.
 */
qio_error( str, stat )
char *str;
int stat;
{

	vmserrno = stat;
	vmsperror(str);
	exit(1);
}

/* io_error - this routine is called if we find an error in the iosb
 * at the completion of some i/o operation.  We print an error msg and
 * return.
 */
io_error( str, stat )
char *str;
u_short stat;
{

	if( stat != SS$_LINKABORT ) {
		vmserrno = stat;
		vmsperror(str);
	}
}
