#ifndef lint
static char rcsid[] =
    "@(#) $Header: raw_packet.c,v 1.9 86/08/01 06:22:35 van Exp $ LBL";
#endif

/*
 * raw packet interface driver and protocol routines
 *
 *	This allows processes on the same machine to get IP packets
 *	to transmit over things like TTY lines.
 *
 * $Log:	raw_packet.c,v $
 * Revision 1.9  86/08/01  06:22:35  van
 * wasn't setting IFF_RUNNING; lint;
 * 
 * Revision 1.8  86/07/30  02:20:07  leres
 * (leres && van) Various changes necessary for 4.3
 * 
 */

#include "param.h"
#include "systm.h"
#include "mbuf.h"
#include "socket.h"
#include "errno.h"
#include "ioctl.h"

#include "../net/if.h"
#include "../net/route.h"
#include "../net/netisr.h"

#include "protosw.h"
#include "socketvar.h"
#include "domain.h"
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"

#ifdef vax
#include "../vax/mtpr.h"
#endif

#define	RPMTU	(1024+512)	/* (same as loopback & ethernet) */

static struct sockaddr_in rpdst = { AF_RAWPACKET };
static struct sockproto rpproto = { PF_RAWPACKET };

struct	ifnet rpif;
int	rpoutput(), rpioctl(), rpacket_output(), raw_usrreq();

extern	struct domain rawpacketdomain;

struct protosw rawpacketsw[] = {
{ SOCK_RAW,	&rawpacketdomain,	0,		PR_ATOMIC|PR_ADDR,
  0,		rpacket_output,		0,		0,
  raw_usrreq,
  0,		0,			0,		0},
};

struct domain rawpacketdomain =
    { AF_RAWPACKET, "rawpacket", 0, 0, 0,
      rawpacketsw, &rawpacketsw[sizeof(rawpacketsw)/sizeof(rawpacketsw[0])] };

/*
 *	Attach the raw packet pseudo-device
 */
rpattach()
{
	register struct ifnet *ifp = &rpif;

	ifp->if_unit = 0;
	ifp->if_name = "rp";
	ifp->if_mtu = RPMTU;
	ifp->if_ioctl = rpioctl;
	ifp->if_output = rpoutput;
	ifp->if_flags = 0;
	if_attach(ifp);
}

/*
 * Raw Packet output:  Send the packet back via the RAWPACKET protocol
 *
 * The unmodified ip packet is hung on the raw socket input queue.  The
 * thing that picks up this packet will presumably have to know the
 * immediate destination (as opposed to the final destination that's in
 * the ip packet).  The right way to supply this info is probably to add a
 * header that contains the immediate source & destination addresses.
 * Rather than do the right thing, I took advantage of the fact that the
 * "src" address is queued with the data.  The addresses passed to
 * "raw_input" are arranged so that the immediate destination address
 * becomes the "from" address when you do a "recvfrom".  This is crufty &
 * will break any attempts to use rp "connections" (as opposed to
 * datagrams) but it saves a little memory & a few cpu cycles.
 */
rpoutput(ifp, m0, dst)
	struct ifnet *ifp;
	struct mbuf *m0;
	struct sockaddr *dst;
{
	int s = splimp();
	int error = 0;

	ifp->if_opackets++;
	switch (dst->sa_family) {

#ifdef INET
	case AF_INET:
#ifdef notdef
		/*
		 * the code should really look like this but, until
		 * Berkeley fixes the raw input routines, we can't
		 * find out about q overflows.
		 */
		if (error = raw_input(m0, &rpproto, dst,
					(struct sockaddr *)&rpdst))
			ifp->if_collisions++;
#else notdef
		raw_input(m0, &rpproto, dst, (struct sockaddr *)&rpdst);
#endif notdef
		break;
#endif
	default:
		m_freem(m0);
		splx(s);
		ifp->if_oerrors++;
		printf("rp%d: can't handle af%d\n", ifp->if_unit,
			dst->sa_family);
		return (EAFNOSUPPORT);
	}
	splx(s);
	return (error);
}


/*
 *	Raw Packet input:  Send the packet back through IP
 */
/*ARGSUSED*/
rpacket_output(m0, dummy)
	register struct mbuf *m0;
	struct socket *dummy;
{
	register struct ifnet *ifp = &rpif;
	register struct ifqueue *ifq;
	register int s;
	register struct mbuf *m;
	register int error = 0;

	ifp->if_ipackets++;
	if (! (ifp->if_flags & IFF_UP)) {
		error = ENETDOWN;
		goto bad;
	}
	/*
	 * Place interface pointer before the data
	 * for the receiving protocol.
	 */
	if (m0->m_off <= MMAXOFF &&
	    m0->m_off >= MMINOFF + sizeof(struct ifnet *)) {
		m0->m_off -= sizeof(struct ifnet *);
		m0->m_len += sizeof(struct ifnet *);
	} else {
		MGET(m, M_DONTWAIT, MT_HEADER);
		if (m == (struct mbuf *)0) {
			error = ENOBUFS;
			goto bad;
		}
		m->m_off = MMINOFF;
		m->m_len = sizeof(struct ifnet *);
		m->m_next = m0;
		m0 = m;
	}
	*(mtod(m0, struct ifnet **)) = ifp;

	ifq = &ipintrq;
	s = splimp();
	if (IF_QFULL(ifq)) {
		IF_DROP(ifq);
		error = ENOBUFS;
		goto bad;
	}
	IF_ENQUEUE(ifq, m0);
	schednetisr(NETISR_IP);
	splx(s);
	return(0);
bad:
	m_freem(m0);
	splx(s);
	ifp->if_ierrors++;
	return (error);
}


/*
 * Process an ioctl request.
 */
rpioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	int cmd;
	caddr_t data;
{
	register struct ifaddr *ifa = (struct ifaddr *)data;
	int s = splimp(), error = EINVAL;

	switch (cmd) {

	case SIOCSIFADDR:
		switch (ifa->ifa_addr.sa_family) {
#ifdef INET
		case AF_INET:
			ifp->if_flags |= (IFF_UP|IFF_RUNNING);
			/* remember our inet address */
			rpdst.sin_addr = IA_SIN(ifa)->sin_addr;
			error = 0;
			break;
#endif
		}
		break;
	}
	splx(s);
	return (error);
}
